;-----------------------------------------------------------
; Y7SETUP version 0.2a
; y7setup.asm
; Copyrigt (C) 2022-2023 あめさろだ (rainy-srd, r-srd)
; Released under the 3-clause BSD License.
;-----------------------------------------------------------

; このファイルで定義されている関数は特記がない限り以下の仕様
; * ax レジスタ以外は破壊しない
; * 32bit レジスタの上位 16bit は破壊されうる

; 486 かつ PCI 搭載のマザーボードは存在する(PCI/I-486SV2 等)
; そのような製品の PCI BIOS の挙動はわからないが一応 .486 としておく
.486

INCLUDE vpmxfer.inc


VENDOR_ID_YAMAHA equ 1073h
DEVICE_ID_YMF724 equ 000dh
DEVICE_ID_YMF744 equ 0010h
DEVICE_ID_YMF754 equ 0012h

; プログラムの動作モード
CMD_LIST	equ 0
CMD_SET		equ 1
CMD_GET		equ 2
CMD_HELP	equ 0feh
CMD_UNKNOWN	equ 0ffh
N_COMMANDS_COMP	equ 3		; compare_string_list で判定されるコマンド数

; set コマンドの設定対象項目
SITM_OPL3	equ 0
SITM_SBPRO	equ 1
SITM_LVOL	equ 2		; YMF7x4 legacy audio output
SITM_AVOL	equ 3		; AC'97 master volume
SITM_IVOL	equ 4		; AC'97 line in volume
N_SETTING_ITEMS	equ 5
; デバイスタイプ
DTYP_YMF724	equ 0
DTYP_YMF744	equ 1
DTYP_YMF754	equ 2
DTYP_OTHER	equ 0ffh	; 724, 744, 754 以外

; PCI コンフィギュレーション空間のレジスタ
CONF_REG_BAR0		equ 10h
CONF_REG_LAC		equ 40h		; legacy audio control
CONF_REG_EXTENDED_LAC	equ 42h		; extended legacy audio control
CONF_REG_DS_1_CTRL	equ 48h		; ds-1 control
CONF_REG_FM_BASE	equ 60h		; fm synthesizer base address
CONF_REG_SBPRO_BASE	equ 62h		; sound blaster base address
; MMIO のレジスタ
MMIO_REG_AC97_CMD_DATA	equ 60h		; command data
MMIO_REG_AC97_CMD_ADDR	equ 62h		; command address
MMIO_REG_AC97_P_ST_DATA	equ 64h		; primary status data
MMIO_REG_AC97_P_ST_ADDR	equ 66h		; primary status address
MMIO_REG_LVOL		equ 80h		; legacy audio output volume
MMIO_AC97_WRITE		equ 0000h	; for command address
MMIO_AC97_READ		equ 8000h
; AC'97 のレジスタ
AC97_REG_MASTER_VOL	equ 02h		; master volume
AC97_REG_LINE_IN_VOL	equ 10h		; line in volume
AC97_REG_PCM_OUT_VOL	equ 18h		; PCM output volume

; set, get コマンドのエラー
ERR_NO_ERROR		equ 0
ERR_NO_ARGS		equ 1	; 引数がない
ERR_INVALID_POS		equ 2	; デバイス位置指定がおかしい
ERR_UNEXP_DEV		equ 3	; 指定の位置に指定のデバイスがない
ERR_INVALID_ITEM	equ 4	; 設定項目指定がおかしい
ERR_NO_SETTING_ARGS	equ 5	; 設定項目の引数がない
ERR_READ_PCI_CONFIG	equ 6	; pci コンフィギュレーション空間の read 失敗
ERR_OPL3_ADDR		equ 7	; opl3 のアドレス指定がおかしい
ERR_WRITE_PCI_CONFIG	equ 8	; pci コンフィギュレーション空間の write 失敗
ERR_SBPRO_ADDR		equ 9	; sbpro のアドレス指定がおかしい
ERR_NO_VCPI		equ 10	; VCPI がない
ERR_INIT_VCPI		equ 11	; VCPI 初期化失敗
ERR_READ_PCI_MMIO	equ 12	; pci の mmio からの読み出し失敗
ERR_WRITE_PCI_MMIO	equ 13	; pci の mmio への書き込み失敗
ERR_AC97_VOL		equ 14	; AC'97 master volume の指定がおかしい
ERR_LEGACY_VOL		equ 15	; ymf7x4 のレガシオーディオ音量の指定がおかしい
ERR_PBUSY_TIMEOUT	equ 16	; PBUSY の polling でタイムアウトした
ERR_AC97_IVOL		equ 17	; AC'97 の line in volume の指定がおかしい

DGROUP GROUP _DATA
_DATA SEGMENT WORD PUBLIC USE16 'DATA'
; これらの関数の戻値(al)がプログラムの戻値になる
mode_proc_tbl	dw	offset mode_proc_list,
			offset mode_proc_set,
			offset mode_proc_get

device_id_tbl	dw	DEVICE_ID_YMF724,
			DEVICE_ID_YMF744,
			DEVICE_ID_YMF754
ymf724_fmio_tbl	dw	388h, 398h, 3a0h, 3a8h
ymf724_sbio_tbl	dw	220h, 240h, 260h, 280h

msg_err_tbl	dw	offset msg_sgerr_no_args,
			offset msg_sgerr_invalid_pos,
			offset msg_sgerr_unexp_dev,
			offset msg_sgerr_invalid_item,
			offset msg_sgerr_no_set_args,
			offset msg_sgerr_read_pci_conf,
			offset msg_sgerr_opl3_addr,
			offset msg_sgerr_write_pci_conf,
			offset msg_sgerr_sbpro_addr,
			offset msg_sgerr_no_vcpi,
			offset msg_sgerr_init_vcpi,
			offset msg_sgerr_read_pci_mmio,
			offset msg_sgerr_write_pci_mmio,
			offset msg_sgerr_ac97_vol,
			offset msg_sgerr_legacy_vol,
			offset msg_sgerr_pbusy_timeout,
			offset msg_sgerr_ac97_ivol
str_cmd_tbl	dw	offset str_cmd_list,
			offset str_cmd_set,
			offset str_cmd_get
str_item_tbl	dw	offset str_item_opl3,
			offset str_item_sbpro,
			offset str_item_lvol,
			offset str_item_avol,
			offset str_item_ivol
str_ymf_tbl	dw	offset str_ymf_724,
			offset str_ymf_744,
			offset str_ymf_754

bar0		dd ?
; PCI の MMIO との間で転送を行う際のバッファ(4bytes)
mmio_buf	dw ?, ?
; AC'97 の busy flag をポーリングする時に読んだデータを置くバッファ
poll_buf	dw ?
; コマンドパラメータ
; sg_***:  set, get コマンド共通
; set_***: set コマンドのみで使うパラメータ
sg_err_submsg	dw 0	; set コマンドエラー時に表示する補足情報
sg_dev_pos	dw ?	; PCI 上のデバイス位置
set_arg		dw ?	; set コマンドの設定項目の引数が es:[[set_arg]] に存在
set_item	db ?	; 設定対象項目
sg_dev_type	db ?	; コマンド対象のデバイスの種類
sg_err		db ERR_NO_ERROR	; set, get コマンドのエラー

cmd		db CMD_UNKNOWN
vcpi_init_ok	db 0	; VCPI が初期化済みかどうか

; コマンド文字列
str_cmd_list db 'list', 0
str_cmd_set db 'set', 0
str_cmd_get db 'get', 0
; 設定項目文字列
str_item_opl3 db 'opl3', 0
str_item_sbpro db 'sbpro', 0
str_item_lvol db 'lvol', 0
str_item_avol db 'avol', 0
str_item_ivol db 'ivol', 0
; チップ名文字列
str_ymf_724 db 'YMF724', 0
str_ymf_744 db 'YMF744', 0
str_ymf_754 db 'YMF754', 0

msg_start db 'Y7SETUP v0.2a (2023/11/03), YMF724/744/754 setup utility',
	0dh, 0ah, 0
msg_err_header db 'Error: ', 0
msg_err_no_pci_bios db 'PCI BIOS not found.', 0dh, 0ah, 0
msg_err_unknown_cmd db 'Unknown command. Run "Y7SETUP.EXE help" for help.',
			0dh, 0ah, 0
msg_serr_common db 'set command failed.', 0dh, 0ah, 0
msg_gerr_common db 'get command failed.', 0dh, 0ah, 0
msg_sgerr_no_args db 'No option is specified.', 0dh, 0ah, 0
msg_sgerr_invalid_pos db 'Device position is invalid or not specified.',
			0dh, 0ah, 0
msg_sgerr_unexp_dev db 'YMF724, 744 or 754 not found at specifed position.',
			0dh, 0ah, 0
msg_sgerr_invalid_item db 'Unexpected setting item',
			' or setting item is not specified.', 0dh, 0ah, 0
msg_sgerr_no_set_args db 'Option(s) of setting item are not specified.',
			 0dh, 0ah, 0
msg_sgerr_read_pci_conf db 'Failed to read PCI configuration space.',
			0dh, 0ah, 0
msg_sgerr_opl3_addr db 'OPL3 base address is invalid.', 0dh, 0ah, 0
msg_sgerr_write_pci_conf db 'Failed to write PCI configuration space.',
			0dh, 0ah, 0
msg_sgerr_sbpro_addr db 'SB Pro base address is invalid.', 0dh, 0ah, 0
msg_sgerr_no_vcpi db 'VCPI not found.', 0dh, 0ah, 0
msg_sgerr_init_vcpi db 'Failed to initialize VCPI.', 0dh, 0ah, 0
msg_sgerr_read_pci_mmio db 'Failed to read PCI MMIO.', 0dh, 0ah, 0
msg_sgerr_write_pci_mmio db 'Failed to write data to PCI MIMO.', 0dh, 0ah, 0
msg_sgerr_ac97_vol db "Master volume of AC'97 codec is invalid.", 0dh, 0ah, 0
msg_sgerr_legacy_vol db 'Volume of YMF7x4 legacy audio block is invalid.',
			0dh, 0ah, 0
msg_sgerr_pbusy_timeout db "Failed to communicate with AC'97",
			' (CRST is set?).', 0dh, 0ah, 0
msg_sgerr_ac97_ivol db "Line in volume of AC'97 codec is invalid.", 0dh, 0ah, 0
msg_724_opl3_err db 'Valid address for YMF724 are 388, 398, 3a0 or 3a8.',
			0dh, 0ah, 0
msg_744_754_opl3_err db 'Address must be 4 bytes aligned on YMF744/754.',
			0dh, 0ah, 0
msg_724_sbpro_err db 'Valid address for YMF724 are 220, 240, 260, 280.',
			0dh, 0ah, 0
msg_744_754_sbpro_err db 'Address must be 16 bytes aligned on YMF744/754.',
			0dh, 0ah, 0
msg_found db ' found.', 0dh, 0ah, 0
msg_LAD    db 'LAD (Legacy Audio Disable)      : ', 0
msg_FMEN   db 'FMEN (FM Synthesizer Enable)    : ', 0
msg_SBEN   db 'SBEN (Sound Blaser Enable)      : ', 0
msg_FMBase db 'FM Synthesizer Base Address     : ', 0
msg_SBBase db 'Sound Blaster Base Address      : ', 0
msg_CRST   db "CRST (AC'97 reset signal)       : ", 0
msg_lvol   db 'YMF7x4 legacy audio volume (L/R): ', 0
msg_amute  db "AC'97 master volume mute        : ", 0
msg_avol   db "AC'97 master volume (L/R)       : ", 0
msg_imute  db "AC'97 line in volume mute       : ", 0
msg_ivol   db "AC'97 line in volume (L/R)      : ", 0
msg_success db 'Operation completed successfully.', 0dh, 0ah, 0
msg_help db 'USAGE: Y7SETUP.EXE [command] (options...)', 0dh, 0ah,
'[command]:', 0dh, 0ah,
'  help    print this help', 0dh, 0ah,
'  list    list up YMF7x4 devices', 0dh, 0ah,
'  get [device pos]    print device settings and status', 0dh, 0ah,
'  set [device pos] [setting item] (options...)', 0dh, 0ah,
'    [device pos  ]:  [bus:device.func] (hex), ex. 00:0D.0', 0dh, 0ah
db '    [setting item]:', 0dh, 0ah,	; line too long のエラー回避
'      opl3     set IO base address of OPL3 block (0-ffff, off)', 0dh, 0ah,
'      sbpro    set IO base address of SB Pro block (0-ffff, off)', 0dh, 0ah,
'      lvol     set YMF7x4 legacy audio block output volume (0-3fff)', 0dh, 0ah,
"      avol     set master volume of AC'97 codec (0-3f)", 0dh, 0ah,
"      ivol     set line in volume of AC'97 codec (0-3f, mute)", 0dh, 0ah,
	0
_DATA ENDS


_TEXT SEGMENT PUBLIC USE16 'CODE'
	assume cs:_TEXT, ds:DGROUP, ss:STACK

start PROC PRIVATE
	mov	ax, DGROUP
	mov	ds, ax
	mov	bx, ss
	sub	bx, ax
	shl	bx, 1
	shl	bx, 1
	shl	bx, 1
	shl	bx, 1
	cli
	mov	ss, ax
	add	sp, bx
	sti

	mov	ax, offset msg_start
	call	putstr

	call	check_pci_bios
	test	ax, ax
	jz	err_no_pci_bios

	call	parse_args
	mov	bl, [cmd]
	cmp	bl, CMD_UNKNOWN
	jz	err_unknown_cmd
	cmp	bl, CMD_HELP
	jz	print_help
	; unknown, help 以外ならエラー確認後テーブルジャンプ
	cmp	[sg_err], ERR_NO_ERROR
	jz	@F
	call	print_sg_err
	mov	al, 1
	jmp	program_end
@@:	xor	bh, bh
	add	bx, bx			; word access
	call	[mode_proc_tbl + bx]

	test	al, al
	jnz	program_end
	mov	dl, al			; save exit code
	mov	ax, offset msg_success
	call	putstr
	mov	al, dl			; restore exit code

	; VCPI 初期化済みなら後始末
.if([vcpi_init_ok] != 0)
	push	ax
	invoke	dispose_vcpi
	pop	ax
.endif
	; exit code は mode_proc_tbl の関数の値をそのまま横流し
program_end:
	mov	ah, 4ch
	int	21h
err_end:
	mov	al, 1
	jmp	program_end
print_help:
	mov	ax, offset msg_help
	call	putstr
	jmp	err_end
err_no_pci_bios:
	mov	ax, offset msg_err_no_pci_bios
	call	print_err
	jmp	err_end
err_unknown_cmd:
	mov	ax, offset msg_err_unknown_cmd
	call	print_err
	jmp	err_end
start ENDP

; エラー処理共通化 (al=1 で return)
; * PCI コンフィギュレーション空間の読み書き失敗時
pci_read_failed:
	mov	[sg_err], ERR_READ_PCI_CONFIG
@@:	call	print_sg_err
	mov	al, 1
	ret
pci_write_failed:
	mov	[sg_err], ERR_WRITE_PCI_CONFIG
	jmp	@B
; * PCI の MMIO の読み書き失敗時
mmio_read_failed:
	mov	[sg_err], ERR_READ_PCI_MMIO
	jmp	@B
mmio_write_failed:
	mov	[sg_err], ERR_WRITE_PCI_MMIO
	jmp	@B
; * PBUSY の polling のタイムアウト時
pbusy_timeout:
	mov	[sg_err], ERR_PBUSY_TIMEOUT
	jmp	@B


; set コマンドの処理開始地点
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; [sg_err] 等が設定されている必要がある
; すべてのレジスタは破壊されうる
mode_proc_set PROC ;PRIVATE
	;--- 指定位置にあるデバイスが何か判定
	mov	ax, [sg_dev_pos]
	call	detect_ymf_type
	jc	pci_read_failed
.if(al == DTYP_OTHER)
	mov	[sg_err], ERR_UNEXP_DEV
	call	print_sg_err
fail:	mov	al, 1
	ret
.endif
	mov	[sg_dev_type], al
	;--- 設定項目で分岐
	mov	al, [set_item]
.if((al == SITM_OPL3) || (al == SITM_SBPRO))
	call	set_cmd_opl3_sbpro
.elseif(al == SITM_LVOL)
	call	set_cmd_lvol
.elseif(al == SITM_AVOL)
	call	set_cmd_avol
.elseif(al == SITM_IVOL)
	call	set_cmd_ivol
.endif
	ret			; set_cmd_*** の戻値を返す
mode_proc_set ENDP


; OPL3/SBPro のベースアドレス設定処理
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; すべてのレジスタは破壊されうる
set_cmd_opl3_sbpro PROC ;PRIVATE
	;--- 「off」が指定されているかチェック
	mov	ax, [set_arg]
	call	check_off_str
	jc	@F
	call	set_cmd_opl3_sbpro_off
	ret			; set_cmd_opl3_sbpro_off の戻値を返す
	;--- 1～4桁の16進数の値があると想定してパース
@@:	mov	ax, [set_arg]
	mov	dl, 4
	call	parse_hex_str_es
	jc	ERR0
.if([set_item] == SITM_OPL3)
	call	set_cmd_opl3_addr
.else ;if([set_item] == SITM_SBPRO)
	call	set_cmd_sbpro_addr
.endif
	ret			; 戻値をそのまま返す
ERR0:
.if([set_item] == SITM_OPL3)
	mov	[sg_err], ERR_OPL3_ADDR
.else
	mov	[sg_err], ERR_SBPRO_ADDR
.endif
	call	print_sg_err
	mov	al, 1
	ret
set_cmd_opl3_sbpro ENDP


; OPL3 のベースアドレスを指定し、それらの機能を有効化する
; 引数: ax=base addr
; 戻値: al=成功?(0: 成功, 非0: 失敗)
; すべてのレジスタは破壊されうる
set_cmd_opl3_addr PROC ;PRIVATE
	cmp	[sg_dev_type], DTYP_YMF724
	jnz	not_724
	;--- YMF724 (アドレスの選択肢に制約がある)
	xor	bx, bx
L1:	cmp	ax, ymf724_fmio_tbl[bx]
	jz	@F
	add	bx, 2		; word access
	cmp	bx, 4 * 2
	jae	invalid_addr
	jmp	L1
invalid_addr:
	mov	[sg_err], ERR_OPL3_ADDR
	mov	[sg_err_submsg], offset msg_724_opl3_err
	call	print_sg_err
fail:	mov	al, 1
	ret
@@:	shr	bx, 1		; FMIO の値にする
	; Extended Legacy Audio Control で設定
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_EXTENDED_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	and	ax, 0fffch
	or	ax, bx			; set FMIO
	or	ax, 10h			; MAIM=1
	mov	cx, ax
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word	; dx=CONF_REG_ELAC
	jc	pci_write_failed
	jmp	set_LAC
	;--- YMF744, 754 は 4byte アライメントなアドレスを任意に指定可能
not_724:
	test	ax, 11b
	jz	@F
	mov	[sg_err], ERR_OPL3_ADDR
	mov	[sg_err_submsg], offset msg_744_754_opl3_err
	call	print_sg_err
	jmp	fail
@@:	; FM Synthesizer Base Address に設定
	mov	cx, ax
	mov	dx, CONF_REG_FM_BASE
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word
	jc	pci_write_failed
	;--- Legacy Audio Control 設定
set_LAC:
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	or	ax, 4002h		; FMEN=1, SIEN=1
	and	ax, 7fd3h		; LAD=0, I/O=0, GPEN=0, MEN=0
	mov	cx, ax
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word	; dx=40h
	jc	pci_write_failed
	xor	al, al
	ret
set_cmd_opl3_addr ENDP
; SB Pro のベースアドレスを指定し、それらの機能を有効化する
; set_cmd_opl3_addr と仕様は同じ
set_cmd_sbpro_addr PROC ;PRIVATE
	cmp	[sg_dev_type], DTYP_YMF724
	jnz	not_724
	;--- YMF724 の場合
	xor	bx, bx
L1:	cmp	ax, ymf724_sbio_tbl[bx]
	jz	@F
	add	bx, 2		; word access
	cmp	bx, 4 * 2
	jae	invalid_addr
	jmp	L1
invalid_addr:
	mov	[sg_err], ERR_SBPRO_ADDR
	mov	[sg_err_submsg], offset msg_724_sbpro_err
	call	print_sg_err
fail:	mov	al, 1
	ret
@@:	shr	bx, 1		; SBIO の値にする
	; Extended Legacy Audio Control で設定
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_EXTENDED_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	and	ax, 0fff3h
	or	ax, bx		; set SBIO
	or	ax, 10h		; MAIM=1
	mov	cx, ax
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word	; dx=42h
	jc	pci_write_failed
	jmp	set_LAC
	;--- 744, 754 は 16byte アライメントなアドレスを任意に指定可能
not_724:
	test	ax, 1111b
	jz	@F
	mov	[sg_err], ERR_SBPRO_ADDR
	mov	[sg_err_submsg], offset msg_744_754_sbpro_err
	call	print_sg_err
	jmp	fail
@@:	; Sound Blaster Base Address に設定
	mov	cx, ax
	mov	dx, CONF_REG_SBPRO_BASE
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word
	jc	pci_write_failed
	;--- Legacy Audio Control 設定
set_LAC:
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	or	ax, 4001h		; SBEN=1, SIEN=1
	and	ax, 7fd3h		; LAD=0, I/O=0, GPEN=0, MEN=0
	mov	cx, ax
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word	; dx=40h
	jc	pci_write_failed
	xor	al, al
	ret
set_cmd_sbpro_addr ENDP


; OPL3/SBPro を off にする
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; すべてのレジスタは破壊されうる
; 片方が on である可能性があるので LAD(bit 15) には触らない
set_cmd_opl3_sbpro_off PROC PRIVATE
	;--- 現在の値を読む
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	;--- ビット設定
.if([set_item] == SITM_OPL3)
	and	ax, 0fffdh		; FMEN=0
.else ;if([set_item] == SITM_SBPRO)
	and	ax, 0fffeh		; SBEN=0
.endif
	;--- 書き込み
	mov	cx, ax
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word	; dx=CONF_REG_LAC
	jc	pci_write_failed
	;---
	xor	al, al
	ret
set_cmd_opl3_sbpro_off ENDP


; YMF7x4 のレガシオーディオブロックの音量設定処理
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; すべてのレジスタは破壊されうる
set_cmd_lvol PROC ;PRIVATE
	;--- 音量取得
	mov	ax, [set_arg]
	mov	dl, 4
	call	parse_hex_str_es
	jc	invalid_lvol
	cmp	ax, 3fffh
	ja	invalid_lvol
	mov	[mmio_buf + 0], ax		; save volume
	mov	[mmio_buf + 2], ax
	;--- VCPI 準備
	call	check_and_init_vcpi
	jnc	@F
	mov	[sg_err], al
	call	print_sg_err
	jmp	fail
	;--- 音量書き込み
@@:	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_BAR0
	call	read_pci_conf_dword
	jc	pci_read_failed
	mov	edx, eax
	add	edx, MMIO_REG_LVOL
	mov	ax, offset mmio_buf
	call	get_linear_addr
	invoke	vpmxfer_2, edx, eax, 2
.if(ax != 0)
	jmp	mmio_write_failed
.endif
	xor	al, al
	ret
invalid_lvol:
	mov	[sg_err], ERR_LEGACY_VOL
	call	print_sg_err
fail:	mov	al, 1
	ret
set_cmd_lvol ENDP


; AC'97(AC-2) コーデックの出力音量設定処理
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; すべてのレジスタは破壊されうる
set_cmd_avol PROC PRIVATE
	;--- 音量取得
	mov	ax, [set_arg]
	mov	dl, 2
	call	parse_hex_str_es
	jc	invalid_avol
	cmp	ax, 3fh
	ja	invalid_avol
	mov	di, ax			; save volume
	;--- write PCM Out Vol = 0808h (0db gain)
	mov	[mmio_buf], (MMIO_AC97_WRITE OR AC97_REG_PCM_OUT_VOL)
	mov	ax, 0808h
	mov	[mmio_buf + 2], ax
	call	set_ac97_val
	test	al, al
	jnz	fail
	;--- polling
	call	poll_pbusy_flag
	jc	pbusy_timeout
	;--- write Master Volume = specified value
	mov	[mmio_buf], (MMIO_AC97_WRITE OR AC97_REG_MASTER_VOL)
	mov	ax, di			; restore volume
	mov	ah, al			; (L/R)
	mov	[mmio_buf + 2], ax
	call	set_ac97_val
	ret			; set_ac97_val の戻り値をそのまま返す
invalid_avol:
	mov	[sg_err], ERR_AC97_VOL
	call	print_sg_err
fail:	mov	al, 1
	ret
set_cmd_avol ENDP


; AC'97(AC-2) コーデックの line in 音量設定処理
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; すべてのレジスタは破壊されうる
set_cmd_ivol PROC
	;--- 「mute」が指定されているかチェック
	mov	ax, [set_arg]
	call	check_mute_str
	jc	@F
	mov	ax, 0888h		; mute 時の設定値 (little endian)
	jmp	Lwrite
	;--- 2桁の16進数の値があると想定してパース
@@:	mov	ax, [set_arg]
	mov	dl, 2
	call	parse_hex_str_es
	jc	invalid_ivol
	cmp	ax, 3fh
	ja	invalid_ivol
	;--- AC'97 書き込み
Lwrite:	mov	[mmio_buf], (MMIO_AC97_WRITE OR AC97_REG_LINE_IN_VOL)
	mov	ah, al			; (L/R)
	mov	[mmio_buf + 2], ax
	call	set_ac97_val
	ret			; set_ac97_val の戻り値をそのまま返す
invalid_ivol:
	mov	[sg_err], ERR_AC97_IVOL
	call	print_sg_err
	mov	al, 1
	ret
set_cmd_ivol ENDP


; AC'97 に何らかの 16bit 値を設定する処理の共通化
; 引数: mmio_buf[0-1]=AC'97 送信先アドレス, mmio_buf[2-3]=AC'97 に送るデータ
; 戻値: al=成功?(0: 成功, 1: 失敗)
; 破壊: eax, edx, ecx
set_ac97_val PROC
	;--- CRST=0 にする
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_DS_1_CTRL
	call	read_pci_conf_word
	jc	pci_read_failed
	and	ax, 0fffeh		; CRST=0
	mov	cx, ax
	mov	ax, [sg_dev_pos]
	call	write_pci_conf_word
	jc	pci_write_failed
	;--- VCPI 準備 (未初期化時のみ)
.if ([vcpi_init_ok] == 0)
	call	check_and_init_vcpi
	jnc	@F
	mov	[sg_err], al
	call	print_sg_err
	jmp	fail
.endif
	;--- AC'97 書き込み
	; read BAR0
@@:	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_BAR0
	call	read_pci_conf_dword
	jc	pci_read_failed
	mov	[bar0], eax
	; addr
	mov	ax, offset mmio_buf
	call	get_linear_addr
	mov	edx, [bar0]
	add	edx, MMIO_REG_AC97_CMD_ADDR
	invoke	vpmxfer_2, edx, eax, 1
	test	ax, ax
	jnz	mmio_write_failed
	; data
	mov	ax, offset mmio_buf + 2
	call	get_linear_addr
	mov	edx, [bar0]
	add	edx, MMIO_REG_AC97_CMD_DATA
	invoke	vpmxfer_2, edx, eax, 1
	test	ax, ax
	jnz	mmio_write_failed
	xor	al, al
	ret
fail:	mov	al, 1
	ret
set_ac97_val ENDP


; AC'97 から何らかの 16bit 値を取得する処理の共通化
; 引数: mmio_buf[0-1]=AC'97 読み出しアドレス, [bar0]=BAR0
; 戻値: ax=読んだ値
; 破壊: eax, edx, ecx
; 失敗した場合全部別の失敗時ルーチンへ行くので ret してくれば必ず成功している
get_ac97_val PROC
	mov	ax, offset mmio_buf
	call	get_linear_addr
	mov	edx, [bar0]
	add	edx, MMIO_REG_AC97_CMD_ADDR
	invoke	vpmxfer_2, edx, eax, 1
.if(ax != 0)
	jmp	mmio_write_failed
.endif
	; polling
	call	poll_pbusy_flag
	jc	pbusy_timeout
	; read volume
	mov	ax, offset mmio_buf + 2
	call	get_linear_addr
	mov	edx, [bar0]
	add	edx, MMIO_REG_AC97_P_ST_DATA
	invoke	vpmxfer_2, eax, edx, 1
.if(ax != 0)
	jmp	mmio_read_failed
.endif
	mov	ax, [mmio_buf + 2]
	ret
get_ac97_val ENDP


; 指定された位置の YMF7x4 の設定情報を表示する
; 引数: 無し
; 戻値: al=成功?(0: 成功, 1: 失敗)
; すべてのレジスタは破壊されうる
mode_proc_get PROC PRIVATE
	;--- 指定された位置のデバイス判定
	mov	ax, [sg_dev_pos]
	call	detect_ymf_type
	jc	pci_read_failed
.if(al == DTYP_OTHER)
	mov	[sg_err], ERR_UNEXP_DEV
	call	print_sg_err
fail:	mov	al, 1
	ret
.endif
	mov	[sg_dev_type], al
	;--- デバイスの情報取得と表示
	; LAD, FMEN, SBEN
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	mov	cx, ax			; save result
	mov	dx, cx
	and	dx, 8000h		; LAD
	mov	ax, offset msg_LAD
	call	print_param_one
	mov	dx, cx
	and	dx, 0002h		; FMEN
	mov	ax, offset msg_FMEN
	call	print_param_one
	mov	dx, cx
	and	dx, 0001h		; SBEN
	mov	ax, offset msg_SBEN
	call	print_param_one
	;--- FMBase, SBBase
	cmp	[sg_dev_type], DTYP_YMF724
	jnz	not_724
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_EXTENDED_LAC
	call	read_pci_conf_word
	jc	pci_read_failed
	mov	cx, ax			; save result
	and	ax, 0003h		; FMIO
	add	ax, ax			; word access
	mov	bx, ax
	mov	dx, ymf724_fmio_tbl[bx]
	mov	ax, offset msg_FMBase
	call	print_param_word
	mov	bx, cx
	and	bx, 000ch
	shr	bx, 1			; word access
	mov	dx, ymf724_sbio_tbl[bx]
	mov	ax, offset msg_SBBase
	call	print_param_word
	jmp	L1
not_724:
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_FM_BASE
	call	read_pci_conf_word
	jc	pci_read_failed
	mov	dx, ax
	mov	ax, offset msg_FMBase
	call	print_param_word
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_SBPRO_BASE
	call	read_pci_conf_word
	jc	pci_read_failed
	mov	dx, ax
	mov	ax, offset msg_SBBase
	call	print_param_word
L1:	;--- CRST (一応表示)
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_DS_1_CTRL
	call	read_pci_conf_word
	jc	pci_read_failed
	and	ax, 0001h		; CRST
	mov	dx, ax
	mov	ax, offset msg_CRST
	call	print_param_one
	;--- これ以降 PCI の MMIO にアクセスする必要がある
	call	check_and_init_vcpi
.if(carry?)
	mov	[sg_err], al
	call	print_sg_err
	jmp	fail
.endif
	;--- Legacy Audio Volume
	mov	ax, [sg_dev_pos]
	mov	dx, CONF_REG_BAR0
	call	read_pci_conf_dword		; eax=BAR0
	jc	pci_read_failed
	mov	[bar0], eax
	mov	edx, eax
	add	edx, MMIO_REG_LVOL
	mov	ax, offset mmio_buf
	call	get_linear_addr
	invoke	vpmxfer_2, eax, edx, 2
.if(ax != 0)
	jmp	mmio_read_failed
.endif
	mov	edx, dword ptr mmio_buf		; [31:16]=Rch, [15:0]=Lch
	mov	ax, offset msg_lvol
	call	print_param_words
	;--- AC'97 master volume
	mov	[mmio_buf], (MMIO_AC97_READ OR AC97_REG_MASTER_VOL)
	call	get_ac97_val
	mov	cx, ax
	mov	dx, ax
	and	dx, 8000h		; mute
	mov	ax, offset msg_amute
	call	print_param_one
	mov	dx, cx
	and	dx, 7fffh		; volume
	mov	ax, offset msg_avol
	call	print_param_bytes
	;--- AC'97 line in volume
	mov	[mmio_buf], (MMIO_AC97_READ OR AC97_REG_LINE_IN_VOL)
	call	get_ac97_val
	mov	cx, ax
	mov	dx, ax
	and	dx, 8000h		; mute
	mov	ax, offset msg_imute
	call	print_param_one
	mov	dx, cx
	and	dx, 7fffh		; volume
	mov	ax, offset msg_ivol
	call	print_param_bytes
	;---
	xor	al, al
	ret
mode_proc_get ENDP


; AC'97 の primary codec の PBUSY をポーリングする
; 引数: [bar0] に BAR0 の値が入っている必要がある
; 戻値: CF=結果(1: タイムアウト, 0: PBUSYが0になった)
poll_pbusy_flag PROC PRIVATE
	push	dx
	push	cx	; vpmxfer_2 が __cdecl なので cx も壊れる
	push	di
	;---
	mov	di, 10			; 10 回まで polling する
poll:	mov	ax, offset poll_buf
	call	get_linear_addr
	mov	edx, [bar0]
	add	edx, MMIO_REG_AC97_P_ST_ADDR
	invoke	vpmxfer_2, eax, edx, 1
	test	ax, ax
	jnz	mmio_read_failed
	test	[poll_buf], 8000h	; check PBUSY flag
	jz	@F
	call	wait_some_time
	dec	di
	test	di, di
	jz	timeout
	jmp	poll
	;---
@@:	clc
@@:	pop	di
	pop	cx
	pop	dx
	ret
timeout:
	stc
	jmp	@B
poll_pbusy_flag ENDP


; loop 命令を使って適当な時間待つ
; 引数, 戻値: 無し
wait_some_time PROC PRIVATE
	push	cx
	mov	cx, 0ffffh
@@:	loop	@B
	pop	cx
	ret
wait_some_time ENDP


; YMF7x4 の位置を全部表示する
; 引数: 無し
; 戻値: al=0 (0: 成功, 失敗しない)
mode_proc_list PROC PRIVATE
	; 724
	mov	ax, offset str_ymf_724
	call	putstr
	call	print_colon_space
	mov	ax, DEVICE_ID_YMF724
	call	print_subproc
	; 744
	mov	ax, offset str_ymf_744
	call	putstr
	call	print_colon_space
	mov	ax, DEVICE_ID_YMF744
	call	print_subproc
	; 754
	mov	ax, offset str_ymf_754
	call	putstr
	call	print_colon_space
	mov	ax, DEVICE_ID_YMF754
	call	print_subproc
	xor	al, al
	ret
; 指定された YAMAHA のデバイスの位置を全部表示する
; 引数: ax=デバイスID
; 戻値: 無し
; 「**:**.*, **:**.*, **:**.*」といった表示になる
print_subproc:
	push	dx
	push	cx
	push	bx
	;---
	mov	dx, ax		; save device_id
	xor	cx, cx
listup_loop:
	mov	ax, cx
	mov	bx, dx
	call	find_pci_device
	test	ax, ax
	jz	exit
.if(cx != 0)
	mov	al, ','
	int	29h
	mov	al, ' '
	int	29h
.endif
	mov	ax, bx
	call	print_pci_dev_pos
	inc	cx
	jmp	listup_loop
exit:	mov	al, 0dh
	int	29h
	mov	al, 0ah
	int	29h
	;---
	pop	bx
	pop	cx
	pop	dx
	ret
; 「: 」と表示する
print_colon_space:
	mov	al, ':'
	int	29h
	mov	al, ' '
	int	29h
	ret
mode_proc_list ENDP


; ベンダが YAMAHA の PCI デバイスを探す(デバイス ID の指定も必要)
; 引数: ax=何個目のデバイスを見つけてほしいか, bx=デバイスID
; 戻値: ax=(非0: 発見, 0: 見つからず)
;	bx=位置([15:8|7:3|2:0]=[bus 番号|device 番号|function 番号])
; ax=0 でも bx は破壊される
find_pci_device PROC PRIVATE
	push	cx
	push	dx
	push	si
	;---
	mov	si, ax
	mov	ax, 0b102h
	mov	cx, bx
	mov	dx, VENDOR_ID_YAMAHA
	int	1ah
	test	ah, ah
	jnz	not_found
	mov	ax, 1
	jmp	@F
not_found:
	xor	ax, ax
	;---
@@:	pop	si
	pop	dx
	pop	cx
	ret
find_pci_device ENDP


; 指定した位置のデバイスが YMF724, 744, 754 のどれか調べる
; 引数: ax = 位置
; 戻値: al=結果(DTYP_*), CF=PCI 読み出し成功?(0: 成功, 1: 失敗)
; ah は破壊される
detect_ymf_type PROC PRIVATE
	push	dx
	xor	dx, dx
	call	read_pci_conf_dword
	jc	exit
	cmp	ax, VENDOR_ID_YAMAHA
	jnz	other
	shr	eax, 16
	cmp	ax, DEVICE_ID_YMF724
	jnz	@F
	mov	ax, DTYP_YMF724
	jmp	found
@@:	cmp	ax, DEVICE_ID_YMF744
	jnz	@F
	mov	ax, DTYP_YMF744
	jmp	found
@@:	cmp	ax, DEVICE_ID_YMF754
	jnz	other
	mov	ax, DTYP_YMF754
	jmp	found
other:	mov	ax, DTYP_OTHER
exit:	pop	dx
	ret
	;--- YMF7x4 が見つかった場合、メッセージ表示
found:	
	push	ax			; save result
	push	bx
	mov	bx, ax
	add	bx, bx			; word access
	mov	ax, str_ymf_tbl[bx]
	call	putstr
	mov	ax, offset msg_found
	call	putstr
	pop	bx
	pop	ax
	jmp	exit
detect_ymf_type ENDP


; PCI コンフィギュレーション空間から読む処理の共通部分
; これを直接呼び出すのは想定していないので注意
read_pci_conf_common PROC PRIVATE
	push	bx
	push	di
	;---
	mov	di, dx
	mov	bx, ax
	mov	ax, si
	int	1ah
.if((ah != 0) || carry?)
	stc
.else
	clc
.endif
	;---
	pop	di
	pop	bx
	ret
read_pci_conf_common ENDP
; PCI コンフィギュレーション空間から word で読む
; 引数: ax=デバイスの位置, dx=読むアドレス
; 戻値: ax=読んだ値, CF=成功?(0: 成功, 1: 失敗)
read_pci_conf_word PROC PRIVATE
	push	cx
	push	si
	mov	si, 0b109h
	call	read_pci_conf_common
	mov	ax, cx
	pop	si
	pop	cx
	ret
read_pci_conf_word ENDP
; PCI コンフィギュレーション空間から dword で読む
; 引数: ax=デバイス位置, dx=読むアドレス
; 戻値: eax=読んだ値, CF=成功?(0: 成功, 1: 失敗)
; この関数は cx でなく ecx を保存する
read_pci_conf_dword PROC PRIVATE
	push	ecx
	push	si
	mov	si, 0b10ah
	call	read_pci_conf_common
	mov	eax, ecx
	pop	si
	pop	ecx
	ret
read_pci_conf_dword ENDP


; PCI コンフィギュレーション空間に書く処理の共通部分
; これを直接呼び出すのは想定していないので注意
write_pci_conf_common PROC PRIVATE
	push	bx
	push	di
	;---
	mov	di, dx
	mov	bx, ax
	mov	ax, si
	int	1ah
.if((ah != 0) || carry?)
	stc
.else
	clc
.endif
	;---
	pop	di
	pop	bx
	ret
write_pci_conf_common ENDP
; PCI コンフィギュレーション空間に word で書く
; 引数: ax=デバイス位置, dx=書くアドレス, cx=書く値
; 戻値: CF=成功?(0: 成功, 1: 失敗)
write_pci_conf_word PROC PRIVATE
	push	si
	mov	si, 0b10ch
	call	write_pci_conf_common
	pop	si
	ret
write_pci_conf_word ENDP
; PCI コンフィギュレーション空間に dword で書く
; 引数: ax=デバイス位置, dx=書くアドレス, ecx=書く値
; 戻値: CF=成功?(0: 成功, 1: 失敗)
write_pci_conf_dword PROC PRIVATE
	push	si
	mov	si, 0b10dh
	call	write_pci_conf_common
	pop	si
	ret
write_pci_conf_dword ENDP


; PCI BIOS 存在チェック
; 引数: 無し
; 戻値: ax=PCI BIOS が存在するか(非0=存在する, 0=存在しない)
check_pci_bios PROC PRIVATE
	push	edx
	push	bx
	push	cx
	;---
	xor	ah, ah
	xor	edx, edx
	clc
	mov	ax, 0b101h
	int	1ah		; PCI BIOS 存在確認
.if((ah == 0) || !(carry?)) && (edx == 20494350h)
	mov	ax, 1
.else
	xor	ax, ax
.endif
	;---
	pop	cx
	pop	bx
	pop	edx
	ret
check_pci_bios ENDP


; PCI デバイスの位置表示
; 引数: ax=位置([15:8|7:3|2:0]=[bus 番号|device 番号|function 番号])
; 戻値: 無し
; (BUS):(DEV).(FUNC) で表示する
print_pci_dev_pos PROC PRIVATE
	push	ax
	mov	al, ah
	call	print_al
	mov	al, ':'
	int	29h
	pop	ax
	mov	ah, al		; save AL
	shr	al, 3
	call	print_al
	mov	al, '.'
	int	29h
	mov	al, ah		; restore AL
	and	al, 07h
	jmp	print_al_low
print_pci_dev_pos ENDP


; パラメータ表示補助
; 引数: ds:ax=値の前に表示する文字列, dx=表示したい値
; 戻値: 無し
; dx=0 の時に「0」それ以外は「１」と表示する
print_param_one PROC PRIVATE
	call	putstr
	test	dx, dx
	jz	@F
	mov	al, '1'
	jmp	L1
@@:	mov	al, '0'
L1:	int	29h
	call	print_new_line
	ret
print_param_one ENDP
; 仕様は print_param_one と同じ(dx を表示)
print_param_word PROC PRIVATE
	call	putstr
	mov	ax, dx
	call	print_ax
	mov	al, 'h'			; 16進数のため
	int	29h
	call	print_new_line
	ret
print_param_word ENDP
; 仕様は print_param_one と同じ(edx を [15:0],[31:16]の順で表示)
print_param_words PROC PRIVATE
	call	putstr
	mov	eax, edx
	call	print_ax
	mov	al, 'h'			; 16進数のため
	int	29h
	mov	al, ','
	int	29h
	mov	al, ' '
	int	29h
	shr	eax, 16
	call	print_ax
	mov	al, 'h'
	int	29h
	call	print_new_line
	ret
print_param_words ENDP
; 仕様は print_param_one と同じ(dx を [7:0],[15:8] の順で表示)
print_param_bytes PROC PRIVATE
	call	putstr
	mov	ax, dx
	call	print_al
	mov	al, 'h'			; 16進数のため
	int	29h
	mov	al, ','
	int	29h
	mov	al, ' '
	int	29h
	mov	al, dh
	call	print_al
	mov	al, 'h'
	int	29h
	call	print_new_line
	ret
print_param_bytes ENDP

; エラー表示
; 引数: ds:ax=エラー文字列の先頭アドレス
; 戻値: 無し
print_err PROC PRIVATE
	push	ax
	mov	ax, offset msg_err_header
	call	putstr				; 「ERROR: 」を表示
	pop	ax
	jmp	putstr
print_err ENDP


; set, get コマンドのエラー表示
; 引数: 無し
; 戻値: 無し
print_sg_err PROC PRIVATE
	push	bx
	mov	bl, [sg_err]
	cmp	bl, ERR_NO_ERROR
	jz	exit
	dec	bl
	xor	bh, bh
	add	bx, bx		; word access
	mov	ax, offset msg_err_header
	call	putstr
.if([cmd] == CMD_SET)
	mov	ax, offset msg_serr_common
.else ;if([cmd] == CMD_GET)
	mov	ax, offset msg_gerr_common
.endif
	call	putstr
	mov	al, ' '
	int	29h
	int	29h
	mov	ax, msg_err_tbl[bx]
	call	putstr
	; 補足情報があれば表示
	mov	ax, [sg_err_submsg]
	test	ax, ax
	jz	exit
	mov	bx, ax
	mov	al, ' '
	int	29h
	int	29h
	mov	ax, bx
	call	putstr
exit:	pop	bx
	ret
print_sg_err ENDP


; NULL 終端文字列表示
; 引数: ds:ax=文字列の先頭アドレス
; 戻値: 無し
putstr PROC PRIVATE
	push	si
	mov	si, ax
L1:	mov	al, [si]
	test	al, al
	jz	@F
	int	29h
	inc	si
	jmp	L1
@@:	pop	si
	ret
putstr ENDP


; AX の値を16進数で表示する
; 引数: ax=表示する値
; 戻値: 無し
; この関数が呼ばれても壊れるレジスタはない
print_ax PROC PRIVATE
	push	ax
	mov	al, ah
	call	print_al
	pop	ax
	jmp	print_al
print_ax ENDP


; AL の値を16進数で表示する
; 引数: al=表示する値
; 戻値: 無し
; この関数が呼ばれても壊れるレジスタはない
print_al PROC PRIVATE
	push	ax
	;--- 上位 4bit
	mov	ah, al		; save al
	shr	al, 4
	cmp	al, 09h
	ja	@F
	add	al, '0'
	jmp	L1
@@:	add	al, 'A' - 0ah
L1:	int	29h
	;--- 下位 4bit
	mov	al, ah
	and	al, 0fh
	cmp	al, 09h
	ja	@F
	add	al, '0'
	jmp	L2
@@:	add	al, 'A' - 0ah
L2:	int	29h
	;---
	pop	ax
	ret
print_al ENDP


; AL の下位 4bit を表示する
; 引数: al=表示する値
; 戻値: 無し
; この関数が呼ばれても壊れるレジスタはない
print_al_low PROC PRIVATE
	push	ax
	;--- 下位 4bit
	cmp	al, 09h
	ja	@F
	add	al, '0'
	jmp	L1
@@:	add	al, 'A' - 0ah
L1:	int	29h
	;---
	pop	ax
	ret
print_al_low ENDP


; ax の文字列を16進数としてパースする
; 引数: al=パース対象1文字目, ah=パース対象2文字目
; 戻値: al=パース結果の値, CF=成功したかどうか(1:失敗, 0:成功)
; ah は破壊される
; メモリから mov 直後呼ぶのを想定しているため上位桁と下位桁が逆なので注意
parse_ax PROC PRIVATE
	call	parse_al
	jc	fail
	shl	al, 4
	xchg	ah, al
	call	parse_al
	jc	fail
	or	al, ah
	clc
fail:	ret
parse_ax ENDP


; al の文字を16進数としてパースする
; 引数: al=パース対象の文字列
; 戻値: al=パース結果の値, CF=成功したかどうか(1:失敗, 0:成功)
parse_al PROC PRIVATE
.if(al >= '0' && al <= '9')
	sub	al, '0'
	clc
.elseif(al >= 'a' && al <= 'f')
	sub	al, 'a' - 0ah
	clc
.elseif(al >= 'A' && al <= 'F')
	sub	al, 'A' - 0ah
	clc
.else
	stc
.endif
	ret
parse_al ENDP


; n 桁の16進数表記文字列があると想定してパースする (1 <= n <= 4)
; 引数: es:ax=文字列, dl=パースしたい文字数
; 戻値: ax=パース結果, CF=成功?(1: 失敗, 0: 成功)
; 冒頭の連続する '0' をスキップする機能あり
parse_hex_str_es PROC PRIVATE
.if((dl == 0) || (dl > 4))
	stc
	ret
.endif
	push	dx
	push	cx
	push	bx
	push	si
	push	di
	;---
	mov	si, ax
	xor	cx, cx		; (ch, cl)=(非'0'フラグ, パース完了文字数)
	xor	bx, bx		; パース位置
	xor	di, di		; パース結果
parse_loop:
	mov	al, es:[si + bx]
	call	is_whitespace
	jc	parse_end
.if((al == '0') && (ch == 0))	; 最初の連続する '0' のスキップ
	jmp	skip
.endif
	mov	ch, 1
	cmp	cl, dl
	jae	err		; 指定桁数より多くパースしようとしたらエラー
	call	parse_al
	jc	err
	shl	di, 4
	xor	ah, ah
	or	di, ax
	inc	cl
skip:	inc	bx
	jmp	parse_loop
parse_end:
	mov	ax, di
	clc
	jmp	@F
err:	stc
@@:	;---
	pop	di
	pop	si
	pop	bx
	pop	cx
	pop	dx
	ret
parse_hex_str_es ENDP


; 改行を表示する
; 引数: 無し
; 戻値: 無し
print_new_line PROC PRIVATE
	mov	al, 0dh
	int	29h
	mov	al, 0ah
	int	29h
	ret
print_new_line ENDP


; NULL 終端文字列の長さを求める
; 引数: ds:ax=文字列
; 戻値: ax=文字列の長さ
strlen PROC PRIVATE
	push	si
	mov	si, ax
@@:	cmp	byte ptr [si], 0
	jz	L1
	inc	si
	jmp	@B
L1:	sub	ax, si
	neg	ax
	pop	si
	ret
strlen ENDP


; 空白文字か判定する
; 引数: al=文字
; 戻値: CF=結果(1:空白文字である, 0:空白文字でない)
; どのレジスタも破壊されない
is_whitespace PROC PRIVATE
	cmp	al, ' '
	je	L1
	cmp	al, 09h		; 水平タブ
	je	L1
	cmp	al, 0bh		; 垂直タブ
	je	L1
	cmp	al, 0dh		; CR
	je	L1
	cmp	al, 0dh		; LF
	je	L1
	clc
	ret
L1:	stc
	ret
is_whitespace ENDP


; プログラム引数パース
; 引数: 無し
; 戻値: 無し
; 引数に合わせて cmd とその他変数が設定される
parse_args PROC; PRIVATE
	push	es
	push	bx
	push	di
	;--- パース準備
	mov	ah, 62h
	int	21h		; bx=PSP seg addr
	mov	es, bx
	xor	bx, bx
	;--- コマンド取得
	; 冒頭の空白を飛ばす
	call	skip_whitespace
	test	ax, ax
	jz	exit	; 空白文字だけでコマンド引数末尾に到達した場合
	mov	di, bx
	add	di, 81h		; コマンド開始位置
	call	detect_command
	; CMD_SET, CMD_GET でなければパース終了
	mov	al, [cmd]
.if(al != CMD_SET && al != CMD_GET)
	jnz	exit
.endif
	;--- set, get コマンドの引数パース
	; デバイス位置
	add	bx, 3		; bx="set", "get" の後の空白先頭
	call	skip_whitespace
.if(ax == 0)
	mov	[sg_err], ERR_INVALID_POS
	jmp	exit
.endif
	mov	di, bx
	add	di, 81h		; デバイス位置文字列開始位置
	call	parse_device_pos
.if(ax == 0)
	mov	[sg_err], ERR_INVALID_POS
	jmp	exit
.endif
	cmp	[cmd], CMD_GET
	je	exit		; get ならパースはここまで
	; 設定項目
	add	bx, di		; bx=PCI 位置指定文字列後の空白先頭
	call	skip_whitespace
.if(ax == 0)
	mov	[sg_err], ERR_INVALID_ITEM
	jmp	exit
.endif
	mov	di, bx
	add	di, 81h		; 設定項目文字列開始位置
	call	detect_setting_item
.if(ax == 0)
	mov	[sg_err], ERR_INVALID_ITEM
	jmp	exit
.endif
	; 設定項目の引数
	; 現状全設定項目で引数が1個必要なので存在判定は無条件で行う
	add	bx, di		; bx=設定文字列直後の空白先頭
	call	skip_whitespace
.if(ax == 0)
	mov	[sg_err], ERR_NO_SETTING_ARGS
	jmp	exit
.endif
	add	bx, 81h		; 設定項目引数開始位置
	mov	[set_arg], bx
exit:	pop	di
	pop	bx
	pop	es
	ret

; 空白を飛ばしながら bx を進める(引数終端か非空白文字が出たら終了)
; 引数: es:bx=対象文字列
; 戻値: ax=終了条件(非0=非空白文字が出現, 0=終端に到達)
skip_whitespace:
@@:	cmp	bl, byte ptr es:[80h]
	ja	L99
	mov	al, es:[81h + bx]
	call	is_whitespace
	jnc	L98
	inc	bx
	jmp	@B
L99:	xor	ax, ax
	ret
L98:	mov	ax, 1
	ret
parse_args ENDP


; 設定項目判定
; 引数: es:di=設定項目文字列先頭アドレス
; 戻値: ax=結果(非0: OK, 0: 不明な値), di=設定項目文字列の長さ(OK時)
; [set_item] が設定される
detect_setting_item PROC PRIVATE
	push	cx
	push	bx
	;---
	mov	ax, offset str_item_tbl
	mov	cl, N_SETTING_ITEMS
	call	compare_string_list
	jc	fail
	mov	[set_item], al
	;--- 文字列長計算
	xor	bh, bh
	mov	bl, al
	add	bx, bx		; word access
	mov	ax, str_item_tbl[bx]
	call	strlen
	mov	bx, di		; save di
	mov	di, ax
	; 追加で空白文字存在判定が必要
	mov	al, es:[bx + di]
	call	is_whitespace
	jnc	fail
	mov	ax, 1
	jmp	@F
fail:	xor	ax, ax
@@:	pop	bx
	pop	cx
	ret
detect_setting_item ENDP


; PCI デバイスの位置の引数をパースする
; 引数: es:di=パース対象文字列先頭アドレス
; 戻値: ax=成功?(非0: 成功, 0: 失敗), di=パースされた文字数
; 成功時は [sg_dev_pos] に値が入る
; 受け付ける形式は「**:**.*」(2桁の部分は1桁でも許容する)
parse_device_pos PROC PRIVATE
	push	dx
	push	di
	xor	dx, dx
	;--- バス番号
	mov	ax, word ptr es:[di]
	cmp	ah, ':'
	je	@F				; 1桁?
	cmp	byte ptr es:[di + 2], ':'	; 先に「:」を確認
	jne	fail
	call	parse_ax
	jc	fail
	add	di, 3
	jmp	parse_bus_end
@@:	call	parse_al
	jc	fail
	add	di, 2
parse_bus_end:
	mov	dh, al
	;--- デバイス番号
	mov	ax, word ptr es:[di]
	cmp	ah, '.'
	je	@F
	cmp	byte ptr es:[di + 2], '.'
	jne	fail
	call	parse_ax
	jc	fail
	add	di, 3
	jmp	parse_dev_end
@@:	call	parse_al
	jc	fail
	add	di, 2
parse_dev_end:
	cmp	al, 31
	ja	fail			; デバイス番号は 0～31
	shl	al, 3
	mov	dl, al
	;--- ファンクション番号
	mov	ax, word ptr es:[di]
	xchg	ah, al
	call	is_whitespace
	jnc	fail
	mov	al, ah
	call	parse_al
	cmp	al, 7			; ファンクション番号は 0～7
	ja	fail
	or	dl, al
	inc	di
	;---
	mov	[sg_dev_pos], dx	; 結果保存
	mov	ax, 1
	jmp	@F
fail:	xor	ax, ax
@@:	pop	dx		; もとの di (ax=戻値なので dx を使用)
	sub	di, dx
	pop	dx
	ret
parse_device_pos ENDP


; コマンド判定
; 引数: es:di=コマンド文字列先頭アドレス
; 戻値: 無し (cmd が設定される)
detect_command PROC PRIVATE
	;--- help は特別扱いなので先に判定
	cmp	word ptr es:[di], 'eh'
	jnz	@F
	cmp	word ptr es:[di + 2], 'pl'
	jnz	@F
	mov	al, es:[di + 4]
	call	is_whitespace
	jnc	@F
	mov	[cmd], CMD_HELP
	ret
@@:	;--- help 以外
	push	bx
	push	cx
	mov	ax, offset str_cmd_tbl
	mov	cl, N_COMMANDS_COMP
	call	compare_string_list
	jc	fail
	mov	cl, al		; save al (compare_string_list の結果)
	; 空白文字判定
	xor	bh, bh
	mov	bl, al
	add	bx, bx		; word access
	mov	ax, str_cmd_tbl[bx]
	call	strlen
	mov	bx, ax
	mov	al, es:[di + bx]
	call	is_whitespace
	jnc	fail
	mov	[cmd], cl	; save cl(=al)
fail:	pop	cx
	pop	bx
	ret
detect_command ENDP


; 指定の文字列が文字列候補に含まれているか調べる
; 引数: es:di=指定の文字列のアドレス
;	es:ax=候補文字列(NULL 終端文字列)の配列のアドレス
;	cl=候補文字列の個数
; 戻値: ax=最初に一致した文字列の index
;	CF=一致したものがあるか(1: ない, 0: ある)
; 前方一致でしか見ないので「hoge」と「hoge0」が候補にあるような場合は注意
; 指定の文字列の後ろが空白文字かどうか等もチェックしないので注意
compare_string_list PROC PRIVATE
	test	cl, cl
	jnz	@F
	stc
	ret
	;---
@@:	push	cx
	push	dx
	push	bx
	push	di
	push	si
	push	bp
	;---
	mov	ch, cl		; save cl
	mov	bp, ax		; save ax
list_loop:
	mov	si, ds:[bp]	; このループで比較する候補
	xor	bx, bx
 	compare_loop:
		mov	al, [si + bx]
		test	al, al
		jz	found		; 候補に NULL 文字が出たら発見扱い
		cmp	al, es:[di + bx]
		jnz	compare_end	; 不一致なら次の文字列へ
		inc	bx
		jmp	compare_loop
compare_end:
	add	bp, 2		; word access
	dec	cl
	jnz	list_loop
not_found:
	stc
	jmp	@F
found:
	xor	ah, ah
	mov	al, ch
	sub	al, cl
	clc
	;---
@@:	pop	bp
	pop	si
	pop	di
	pop	bx
	pop	dx
	pop	cx
	ret
compare_string_list ENDP


; es:[ax] からの文字列が ("off" + 空白文字) かチェックする
; 引数: 無し
; 戻値: CF=結果(1: "off" + 空白文字でない, 0: "off" + 空白文字である)
check_off_str PROC PRIVATE
	push	si
	mov	si, ax
	mov	ax, es:[si]
	cmp	ax, 'fo'
	jnz	not_off
	mov	ax, es:[si + 2]
	cmp	al, 'f'
	jnz	not_off
	mov	al, ah
	call	is_whitespace
	jnc	not_off
	clc
	jmp	@F
not_off:
	stc
@@:	pop	si
	ret
check_off_str ENDP
; es:[ax] からの文字列が ("mute" + 空白文字) かチェックする
; 引数: 無し
; 戻値: CF=結果(1: "mute" + 空白文字でない, 0: "mute" + 空白文字である)
check_mute_str PROC
	push	si
	mov	si, ax
	mov	ax, es:[si]
	cmp	ax, "um"
	jnz	not_mute
	mov	ax, es:[si + 2]
	cmp	ax, "et"
	jnz	not_mute
	mov	al, es:[si + 4]
	call	is_whitespace
	jnc	not_mute
	clc
	jmp	@F
not_mute:
	stc
@@:	pop	si
	ret
check_mute_str ENDP


; VCPI の存在チェックと初期化を行う
; 引数: 無し
; 戻値: CF=結果(1: 失敗, 0: 成功), ax=エラーコード
check_and_init_vcpi PROC PRIVATE
	push	ecx		; __cdecl での破壊から保護する
	push	edx
	invoke	check_vcpi
	test	ax, ax
	jz	@F
	mov	ax, ERR_NO_VCPI
	jmp	err
	cmp	[vcpi_init_ok], 0
	jnz	L1		; 既に初期化済みなら初期化処理は飛ばす
@@:	invoke	init_vcpi
	test	ax, ax
	jz	@F
	mov	ax, ERR_INIT_VCPI
	jmp	err
@@:	mov	ax, ERR_NO_ERROR
	mov	[vcpi_init_ok], 1
L1:	clc
	jmp	@F
err:	stc
@@:	pop	edx
	pop	ecx
	ret
check_and_init_vcpi ENDP


; ds:ax をリニアアドレスに変換する
; 引数: ds:ax=変換対象のアドレス
; 戻値: eax=リニアアドレス
get_linear_addr PROC PRIVATE
	push	edx
	xor	edx, edx
	mov	dx, ax
	xor	eax, eax
	mov	ax, ds
	shl	eax, 4
	add	eax, edx
	pop	edx
	ret
get_linear_addr ENDP

_TEXT ENDS


STACK SEGMENT WORD PUBLIC USE16 'STACK'
	org	2048 ; 2KiB
STACK ENDS


END start
